/*
 * shitbot.c - GPIO interface driver for The Shitbot
 *
 * Copyright (C) 2008 Alexandre Boeglin
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>

#ifndef __exit
#define __exit __init
#endif

typedef unsigned char uint8;
typedef unsigned int uint32;

static struct proc_dir_entry* shitbot_proc;

static uint8 gpio_in = 1 << 4;
static uint8 gpio_reset = 1 << 3;

static volatile uint32* gpioaddr_input = (uint32*)0xb8000060;
static volatile uint32* gpioaddr_output = (uint32*)0xb8000064;
static volatile uint32* gpioaddr_enable = (uint32*)0xb8000068;
static volatile uint32* gpioaddr_control = (uint32*)0xb800006c;

static void bit_gpio_set(uint32 mask) {
        uint32 port_state;

        port_state = *gpioaddr_output;
        port_state |= mask;
        *gpioaddr_output = port_state;

        port_state = *gpioaddr_enable;
        port_state &= ~mask;
        *gpioaddr_enable = port_state;
}

static void bit_gpio_clear(uint32 mask) {
        uint32 port_state;

        port_state = *gpioaddr_enable;
        port_state |= mask;
        *gpioaddr_enable = port_state;

        port_state = *gpioaddr_output;
        port_state &= ~mask;
        *gpioaddr_output = port_state;
}
static int bit_gpio_get(int mask) {
        unsigned char port_state;
        port_state = *gpioaddr_input;
        return port_state & mask;
}

static int shitbot_proc_read(char* page, char** start, off_t off, int count, int* eof, void* data)
{
	uint8 status;

	// extra safety, we don't want to return more than one char
	if(off > 0)
		return 0;

	// read gpio_in
	status = bit_gpio_get(gpio_in) ? '0' : '1';
	//printk("shitbot: Read (%c)\n", status);

	// write result to file
	*page = status;
	*eof = 1;
	return 1;
}

static int shitbot_proc_write(struct file *file, const char *buffer, unsigned long count, void *data)
{
	//printk("shitbot: Reset\n");

	// set gpio_reset to 0
	bit_gpio_clear(gpio_reset);

	// sleep long enough for the latch to stabilize
	msleep(1);

	// set gpio_reset back to 1
	bit_gpio_set(gpio_reset);

	return count;
}

static int __init shitbot_init(void)
{
	uint8 gpio_mask;
	printk("shitbot: Init\n");

	// first, procfs
	shitbot_proc = create_proc_entry("shitbot", 0666, NULL);

	if(shitbot_proc == NULL)
	{
		printk("shitbot: Init failed\n");
		remove_proc_entry("shitbot", &proc_root);
		return -EINVAL;
	}

	shitbot_proc->read_proc = shitbot_proc_read;
	shitbot_proc->write_proc = shitbot_proc_write;

	// next, make sure the GPIOs are not masked
	gpio_mask = *gpioaddr_control;
	gpio_mask &= ~(gpio_in | gpio_reset);
	*gpioaddr_control = gpio_mask;

	printk("shitbot: Init succeeded\n");
	return 0;
}

static void __exit shitbot_exit(void)
{
	remove_proc_entry("shitbot", &proc_root);
}

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Alexandre Boeglin");
MODULE_LICENSE("GPL");

module_init(shitbot_init);
module_exit(shitbot_exit);

